#include <stdrsl/ShadingContext.h>
#include <stdrsl/RadianceSample.h>
#include <stdrsl/Math.h>
#include <stdrsl/Colors.h>
#include <stdrsl/OrenNayar.h>
#include <stdrsl/Fresnel.h>

//A simple class-based shader
class diffuseShader(
    //Diffuse
    uniform float roughness = 0.2;
    uniform color diffuse_color = color(1.0);
    uniform string diffuse_map = "";
    uniform float diffuse_intensity = 0.5;
    //Other
    uniform float gamma_correct = 1;
    )
{
  //Class members
  shader m_lights[];
  shader m_material;
  stdrsl_ShadingContext m_context;
  stdrsl_OrenNayar m_diffuse;
  stdrsl_Fresnel m_fresnel;

  varying color diffuse;
  
  //Pipeline methods
  public void construct()
  {
    m_context->construct();
  }

  public void begin()
  {
    m_lights = getlights();
    m_material = this;
    m_context->init();
  }

  public void displacement(output point P; output normal N)
  {
    m_context->m_Nn = normalize(N);
    m_context->reinit();
  }

  public void opacity(output color Oi;)
  {
    //No opacity option implememnted
  }

  //Lighting methods
  public void prelighting(output color Ci, Oi;)
  {
    //Get diffuse colour and initialize diffuse brdf
    diffuse = diffuse_color;
    if(diffuse_map != "")
    {
      diffuse = color texture(diffuse_map);
      if(gamma_correct == 1)
        diffuse = srgbToLinear(diffuse);
    }

    m_diffuse->init(diffuse * diffuse_intensity, roughness);
  }

  public void lighting(output color Ci, Oi;)
  {
    //Calculate direct lighting using material/light methods and multiple importance sampling 
    color direct_diff;
    directlighting(
      m_material,
      m_lights,
      "mis", 1,
      "heuristic", "veachpower2",
      "diffuseresult", direct_diff
      );

    //Calculate sample counts based on ray depth and path tracing attribute
    uniform float idiffSamps, idiffMaxVar;
    m_context->m_SampleMgr->computeIndirectDiffuseSamples(64, 0, idiffSamps, idiffMaxVar);

    //Jitter the surface position to avoid aliasing artifacts on secondary rays
    float sample_base = (m_context->m_RayDepth > 0) ? 0 : 1;

    //Get indirect diffuse contribution using cosine weighted distrobution to account for projected area
    color indirect_diff = indirectdiffuse(
      P,
      m_context->m_Nn,
      idiffSamps, 
      "distribution", "cosine",
      "samplebase", sample_base,
      "maxvariation", idiffMaxVar
      ) * m_diffuse->m_diffColor;

    //Accumilate individual contributions
    Ci = direct_diff + indirect_diff;
  }

  //Sample evaluate and generate methods
  public void evaluateSamples(string distribution; output __radiancesample samples[])
  {
    if(distribution == "diffuse")
    {
      m_diffuse->evalDiffuseSamps(m_context, m_fresnel, samples);
    }
  }

  public void generateSamples(string distribution; output __radiancesample samples[])
  {
    if(distribution == "diffuse")
    {
      m_diffuse->genDiffuseSamps(m_context, m_fresnel, samples);
    }
  }
}